@@ -27,7 +27,72 @@ int qla2x00_update_nvram(scsi_qla_host_t *, EXT_IOCTL *, int);
 int qla2x00_write_nvram_word(scsi_qla_host_t *, uint8_t, uint16_t);
 int qla2x00_send_loopback(scsi_qla_host_t *, EXT_IOCTL *, int);
 int qla2x00_read_option_rom(scsi_qla_host_t *, EXT_IOCTL *, int);
+int qla2x00_read_option_rom_ext(scsi_qla_host_t *, EXT_IOCTL *, int);
 int qla2x00_update_option_rom(scsi_qla_host_t *, EXT_IOCTL *, int);
+int qla2x00_update_option_rom_ext(scsi_qla_host_t *, EXT_IOCTL *, int);
+int qla2x00_get_option_rom_layout(scsi_qla_host_t *, EXT_IOCTL *, int);
+static void qla2x00_get_option_rom_table(scsi_qla_host_t *,
+    INT_OPT_ROM_REGION **, unsigned long * );
+
+/* Option ROM definitions. */
+INT_OPT_ROM_REGION OptionRomTable6312[] = // 128k x20000
+{
+    {INT_OPT_ROM_REGION_ALL,	INT_OPT_ROM_SIZE_2312,
+	    0, INT_OPT_ROM_SIZE_2312-1},
+    {INT_OPT_ROM_REGION_PHBIOS,	0xC400,
+	    0, 0xC400-1},
+    {INT_OPT_ROM_REGION_BCFW,	INT_OPT_ROM_SIZE_2312-0xC400-0x200,
+	    0xC400, INT_OPT_ROM_SIZE_2312-0x200-1},
+    {INT_OPT_ROM_REGION_VPD,	0x200,
+	    INT_OPT_ROM_SIZE_2312-0x200, INT_OPT_ROM_SIZE_2312-1},
+    {INT_OPT_ROM_REGION_NONE,	0,	0,	0}
+};
+// ISP2312 Version 3 Chip.
+INT_OPT_ROM_REGION OptionRomTable6826A[] = // 128k x20000
+{
+    {INT_OPT_ROM_REGION_ALL,	INT_OPT_ROM_SIZE_2312,
+	    0, INT_OPT_ROM_SIZE_2312-1},
+    {INT_OPT_ROM_REGION_PHEFI_PHECFW_PHVPD,	INT_OPT_ROM_SIZE_2312,
+	    0, INT_OPT_ROM_SIZE_2312-1},
+    {INT_OPT_ROM_REGION_NONE,	0,	0,	0}
+};
+
+INT_OPT_ROM_REGION OptionRomTable2312[] = // 128k x20000
+{
+    {INT_OPT_ROM_REGION_ALL,	INT_OPT_ROM_SIZE_2312,
+	    0, INT_OPT_ROM_SIZE_2312-1},
+    {INT_OPT_ROM_REGION_PHBIOS_FCODE_EFI_FW,	INT_OPT_ROM_SIZE_2312,
+	    0, INT_OPT_ROM_SIZE_2312-1},
+    {INT_OPT_ROM_REGION_NONE,	0,	0,	0}
+};
+
+INT_OPT_ROM_REGION OptionRomTable2322[] = // 1 M x100000
+{
+    {INT_OPT_ROM_REGION_ALL, INT_OPT_ROM_SIZE_2322,
+	    0, INT_OPT_ROM_SIZE_2322-1},
+    {INT_OPT_ROM_REGION_PHBIOS_PHFCODE_PHEFI, 0x40000,
+	    0, 0x40000-1},
+    {INT_OPT_ROM_REGION_VPD,	0x200,
+	    0x40000, 0x40000+0x200-1},
+    {INT_OPT_ROM_REGION_FW1,	0x80000,
+	    0x80000, INT_OPT_ROM_SIZE_2322-1},
+    {INT_OPT_ROM_REGION_NONE,	0,	0,	0}
+};
+
+INT_OPT_ROM_REGION OptionRomTable2400[] = // 1 M x100000
+{
+    {INT_OPT_ROM_REGION_ALL,	INT_OPT_ROM_SIZE_2322,
+	    0, INT_OPT_ROM_SIZE_2322-1},
+    {INT_OPT_ROM_REGION_PCI_CFG,	0x100,
+	    0, 0x100-1},
+    {INT_OPT_ROM_REGION_PHBIOS_PHFCODE_PHEFI,	0x40000,
+	    0x100, 0x40000-1},
+    {INT_OPT_ROM_REGION_VPD,	0x200,
+	    0x40000, 0x40000+0x200-1},
+    {INT_OPT_ROM_REGION_FW1,	0x80000,
+	    0x80000, INT_OPT_ROM_SIZE_2322-1},
+    {INT_OPT_ROM_REGION_NONE,	0,	0,	0}
+};
 
 int
 qla2x00_read_nvram(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
@@ -68,7 +133,10 @@ qla2x00_read_nvram(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 
 	/* Dump NVRAM. */
 #if defined(ISP2300)
-	if (ha->device_id == QLA2312_DEVICE_ID) {
+	if (ha->device_id == QLA2312_DEVICE_ID ||
+	    ha->device_id == QLA2322_DEVICE_ID ||
+	    ha->device_id == QLA6312_DEVICE_ID ||
+	    ha->device_id == QLA6322_DEVICE_ID) { 	    
 		data = RD_REG_WORD(&reg->ctrl_status);
 		if ((data >> 14) == 1)
 			base = 0x80;
@@ -106,15 +174,18 @@ qla2x00_read_nvram(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 		wptr++;
  	}
 
-#if defined(ISP2300)
-	if (ha->device_id == QLA2312_DEVICE_ID) {
+#if defined(ISP2300) 
+	if (ha->device_id == QLA2312_DEVICE_ID ||
+	    ha->device_id == QLA2322_DEVICE_ID ||
+	    ha->device_id == QLA6312_DEVICE_ID ||
+	    ha->device_id == QLA6322_DEVICE_ID) { 	    
 		/* Unlock resource */
 		WRT_REG_WORD(&reg->host_semaphore, 0);
+		PCI_POSTING(&reg->host_semaphore);
 	}
 #endif
 
-	ret = copy_to_user((uint8_t *)pext->ResponseAdr, ptmp_buf,
-	    transfer_size * 2);
+	ret = copy_to_user(pext->ResponseAdr, ptmp_buf, transfer_size * 2);
 	if (ret) {
 		pext->Status = EXT_STATUS_COPY_ERR;
 		DEBUG9_10(printk("%s(%ld): inst=%ld ERROR copy rsp buffer.\n",
@@ -149,7 +220,7 @@ qla2x00_read_nvram(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 int
 qla2x00_update_nvram(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 {
-#if defined(ISP2300)
+#if defined(ISP2300) 
 	device_reg_t	*reg = ha->iobase;
 #endif
 #if defined(ISP2100)
@@ -190,17 +261,15 @@ qla2x00_update_nvram(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 	kernel_tmp = (uint8_t *)pnew_nv;
 	usr_tmp = (uint8_t *)pext->RequestAdr;
 
-	ret = verify_area(VERIFY_READ, (void *)usr_tmp, transfer_size);
+	ret = copy_from_user(kernel_tmp, usr_tmp, transfer_size);
 	if (ret) {
 		DEBUG9_10(printk(
-		    "qla2x00_update_nvram: ERROR in buffer verify READ. "
+		    "qla2x00_update_nvram: ERROR in buffer copy READ. "
 		    "RequestAdr=%p\n", pext->RequestAdr);)
 		qla2x00_free_ioctl_scrap_mem(ha);
 		return ret;
 	}
 
-	copy_from_user(kernel_tmp, usr_tmp, transfer_size);
-
 	kernel_tmp = (uint8_t *)pnew_nv;
 
 	/* we need to checksum the nvram */
@@ -215,7 +284,10 @@ qla2x00_update_nvram(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 
 	/* Write to NVRAM */
 #if defined(ISP2300)
-	if (ha->device_id == QLA2312_DEVICE_ID) {
+	if (ha->device_id == QLA2312_DEVICE_ID ||
+	    ha->device_id == QLA2322_DEVICE_ID ||
+	    ha->device_id == QLA6312_DEVICE_ID ||
+	    ha->device_id == QLA6322_DEVICE_ID) { 	    
 		data = RD_REG_WORD(&reg->ctrl_status);
 		if ((data >> 14) == 1)
 			base = 0x80;
@@ -254,9 +326,13 @@ qla2x00_update_nvram(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 	}
 
 #if defined(ISP2300)
-	if (ha->device_id == QLA2312_DEVICE_ID) {
+	if (ha->device_id == QLA2312_DEVICE_ID ||
+	    ha->device_id == QLA2322_DEVICE_ID ||
+	    ha->device_id == QLA6312_DEVICE_ID ||
+	    ha->device_id == QLA6322_DEVICE_ID) { 	    
 		/* Unlock resource */
 		WRT_REG_WORD(&reg->host_semaphore, 0);
+		PCI_POSTING(&reg->host_semaphore);
 	}
 #endif
 
@@ -267,7 +343,12 @@ qla2x00_update_nvram(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 
 	DEBUG9(printk("qla2x00_update_nvram: exiting.\n");)
 
-	return 0;
+	/* Schedule DPC to restart the RISC */
+	set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
+	up(ha->dpc_wait);
+
+	ret = qla2x00_wait_for_hba_online(ha);
+	return ret;
 }
 
 int
@@ -373,20 +454,15 @@ qla2x00_send_loopback(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 		return pext->Status;
 	}
 
-	status = verify_area(VERIFY_READ, (void *)pext->RequestAdr,
-	    pext->RequestLen);
+	status = copy_from_user(&req, pext->RequestAdr, pext->RequestLen);
 	if (status) {
 		pext->Status = EXT_STATUS_COPY_ERR;
-		DEBUG9_10(printk("qla2x00_send_loopback: ERROR verify read of "
+		DEBUG9_10(printk("qla2x00_send_loopback: ERROR copy read of "
 		    "request buffer.\n");)
 		return pext->Status;
 	}
 
-	copy_from_user((uint8_t *)&req, (uint8_t *)pext->RequestAdr,
-	    pext->RequestLen);
-
-	status = verify_area(VERIFY_READ, (void *)pext->ResponseAdr,
-	    pext->ResponseLen);
+	status = copy_from_user(&rsp, pext->ResponseAdr, pext->ResponseLen);
 	if (status) {
 		pext->Status = EXT_STATUS_COPY_ERR;
 		DEBUG9_10(printk("qla2x00_send_loopback: ERROR verify read of "
@@ -394,9 +470,6 @@ qla2x00_send_loopback(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 		return pext->Status;
 	}
 
-	copy_from_user((uint8_t *)&rsp, (uint8_t *)pext->ResponseAdr,
-	    pext->ResponseLen);
-
 	if (req.TransferCount > req.BufferLength ||
 	    req.TransferCount > rsp.BufferLength) {
 
@@ -411,18 +484,15 @@ qla2x00_send_loopback(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 		return pext->Status;
 	}
 
-	status = verify_area(VERIFY_READ, (void *)req.BufferAddress,
+	status = copy_from_user(ha->ioctl_mem, req.BufferAddress,
 	    req.TransferCount);
 	if (status) {
 		pext->Status = EXT_STATUS_COPY_ERR;
-		DEBUG9_10(printk("qla2x00_send_loopback: ERROR verify read of "
+		DEBUG9_10(printk("qla2x00_send_loopback: ERROR copy read of "
 		    "user loopback data buffer.\n");)
 		return pext->Status;
 	}
 
-	copy_from_user((uint8_t *)ha->ioctl_mem, (uint8_t *)req.BufferAddress,
-	    req.TransferCount);
-
 	DEBUG9(printk("qla2x00_send_loopback: req -- bufadr=%p, buflen=%x, "
 	    "xfrcnt=%x, rsp -- bufadr=%p, buflen=%x.\n",
 	    req.BufferAddress, req.BufferLength, req.TransferCount,
@@ -473,22 +543,19 @@ qla2x00_send_loopback(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 		}
 	}
 
-	status = verify_area(VERIFY_WRITE, (void *)rsp.BufferAddress,
+	DEBUG9(printk("qla2x00_send_loopback: loopback mbx cmd ok. "
+	    "copying data.\n");)
+
+	/* put loopback return data in user buffer */
+	status = copy_to_user(rsp.BufferAddress, ha->ioctl_mem,
 	    req.TransferCount);
 	if (status) {
 		pext->Status = EXT_STATUS_COPY_ERR;
 		DEBUG9_10(printk("qla2x00_send_loopback: ERROR verify "
 		    "write of return data buffer.\n");)
-		return status ;
+		return (status);
 	}
 
-	DEBUG9(printk("qla2x00_send_loopback: loopback mbx cmd ok. "
-	    "copying data.\n");)
-
-	/* put loopback return data in user buffer */
-	copy_to_user((uint8_t *)rsp.BufferAddress,
-	    (uint8_t *)ha->ioctl_mem, req.TransferCount);
-
 	rsp.CompletionStatus = ret_mb[0];
 
 	if (ha->current_topology == ISP_CFG_F) {
@@ -504,18 +571,14 @@ qla2x00_send_loopback(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 		}
 	}
 
-	status = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr,
-	    pext->ResponseLen);
+	status = copy_to_user(pext->ResponseAdr, &rsp, pext->ResponseLen);
 	if (status) {
 		pext->Status = EXT_STATUS_COPY_ERR;
-		DEBUG9_10(printk("qla2x00_send_loopback: ERROR verify "
+		DEBUG9_10(printk("qla2x00_send_loopback: ERROR copy "
 		    "write of response buffer.\n");)
 		return pext->Status;
 	}
 
-	copy_to_user((uint8_t *)pext->ResponseAdr, (uint8_t *)&rsp,
-	    pext->ResponseLen);
-
 	pext->Status       = EXT_STATUS_OK;
 	pext->DetailStatus = EXT_STATUS_OK;
 
@@ -524,18 +587,17 @@ qla2x00_send_loopback(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 	return pext->Status;
 }
 
-int qla2x00_read_option_rom(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
+int
+qla2x00_read_option_rom(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 {
 	uint8_t		*usr_tmp;
-	uint32_t	addr;
-	uint32_t	midpoint;
 	uint32_t	transfer_size;
-	uint8_t		data;
-	device_reg_t	*reg = ha->iobase;
-	unsigned long	cpu_flags;
 
 	DEBUG9(printk("%s: entered.\n", __func__);)
 
+	if (pext->SubCode)
+		return qla2x00_read_option_rom_ext(ha, pext, mode);
+
 	if (pext->ResponseLen != FLASH_IMAGE_SIZE) {
 		pext->Status = EXT_STATUS_BUFFER_TOO_SMALL;
 		return (1);
@@ -543,24 +605,11 @@ int qla2x00_read_option_rom(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 
 	transfer_size = FLASH_IMAGE_SIZE;
 
-	midpoint = FLASH_IMAGE_SIZE / 2;
 	usr_tmp = (uint8_t *)pext->ResponseAdr;
 
 	/* Dump FLASH. */
-	spin_lock_irqsave(&ha->hardware_lock, cpu_flags);
-	qla2x00_flash_enable(ha);
-	WRT_REG_WORD(&reg->nvram, 0);
-	for (addr = 0; addr < transfer_size; addr++, usr_tmp++) {
-		if (addr == midpoint)
-			WRT_REG_WORD(&reg->nvram, NV_SELECT);
-
-		data = qla2x00_read_flash_byte(ha, addr);
-		if (addr % 100)
-			udelay(10);
-		__put_user(data, usr_tmp);
-	}
-	qla2x00_flash_disable(ha);
-	spin_unlock_irqrestore(&ha->hardware_lock, cpu_flags);
+ 	qla2x00_update_or_read_flash(ha, usr_tmp, 0, 
+	    FLASH_IMAGE_SIZE, QLA2X00_READ); 
 
 	pext->Status = EXT_STATUS_OK;
 	pext->DetailStatus = EXT_STATUS_OK;
@@ -570,16 +619,75 @@ int qla2x00_read_option_rom(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 	return (0);
 }
 
-int qla2x00_update_option_rom(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
+int
+qla2x00_read_option_rom_ext(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
+{
+	uint8_t		*usr_tmp;
+	int		iter, found;
+	uint32_t	saddr, length;
+
+
+	DEBUG9(printk("%s: entered.\n", __func__);)
+
+	found = 0;
+	saddr = length = 0;
+	/* Retrieve region or raw starting address. */
+	if (pext->SubCode == 0xFFFF) {
+		saddr = pext->Reserved1;
+		length = pext->RequestLen;
+		found++;
+	} else {
+		INT_OPT_ROM_REGION *OptionRomTable = NULL;
+		unsigned long  OptionRomTableSize;
+		/* Pick the right OptionRom table based on device id */
+		qla2x00_get_option_rom_table(ha, &OptionRomTable, 
+		    &OptionRomTableSize);
+		for (iter = 0; OptionRomTable != NULL && iter <
+		    (OptionRomTableSize / sizeof(INT_OPT_ROM_REGION));
+		    iter++) {
+			if (OptionRomTable[iter].Region == pext->SubCode) {
+				saddr = OptionRomTable[iter].Beg;
+				length = OptionRomTable[iter].Size;
+				found++;
+				break;
+			}
+		}
+	}
+	if (!found) {
+		pext->Status = EXT_STATUS_ERR;
+		return 1;
+	}
+	if (pext->ResponseLen < length) {
+		pext->Status = EXT_STATUS_COPY_ERR;
+		return 1;
+	}
+
+	usr_tmp = (uint8_t *)pext->ResponseAdr;
+
+	/* Dump FLASH. */
+ 	qla2x00_update_or_read_flash(ha, usr_tmp, saddr, length, QLA2X00_READ); 
+
+	pext->Status = EXT_STATUS_OK;
+	pext->DetailStatus = EXT_STATUS_OK;
+
+	DEBUG9(printk("%s: exiting.\n", __func__);)
+
+	return 0;
+}
+
+int
+qla2x00_update_option_rom(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 {
 	int		ret;
 	uint8_t		*usr_tmp;
 	uint8_t		*kern_tmp;
 	uint16_t	status;
-	unsigned long	cpu_flags;
 
 	DEBUG9(printk("%s: entered.\n", __func__);)
 
+	if (pext->SubCode)
+		return qla2x00_update_option_rom_ext(ha, pext, mode);
+
 	if (pext->RequestLen != FLASH_IMAGE_SIZE) {
 		pext->Status = EXT_STATUS_COPY_ERR;
 		return (1);
@@ -590,14 +698,6 @@ int qla2x00_update_option_rom(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 
 	/* Read from user buffer */
 	usr_tmp = (uint8_t *)pext->RequestAdr;
-	ret = verify_area(VERIFY_READ, (void *)usr_tmp, FLASH_IMAGE_SIZE);
-	if (ret) {
-		pext->Status = EXT_STATUS_COPY_ERR;
-		DEBUG9_10(printk("%s: ERROR in buffer verify READ. "
-				"RequestAdr=%p\n",
-				__func__, pext->RequestAdr);)
-		return (ret);
-	}
 
 	kern_tmp = (uint8_t *)KMEM_ZALLOC(FLASH_IMAGE_SIZE, 30);
 	if (kern_tmp == NULL) {
@@ -606,16 +706,18 @@ int qla2x00_update_option_rom(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 			"%s: ERROR in flash allocation.\n", __func__);
 		return (1);
 	}
-	copy_from_user(kern_tmp, usr_tmp, FLASH_IMAGE_SIZE);
+	ret = copy_from_user(kern_tmp, usr_tmp, FLASH_IMAGE_SIZE);
+	if (ret) {
+		KMEM_FREE(kern_tmp, FLASH_IMAGE_SIZE);
+		pext->Status = EXT_STATUS_COPY_ERR;
+		DEBUG9_10(printk("%s: ERROR in buffer copy READ. "
+				"RequestAdr=%p\n",
+				__func__, pext->RequestAdr);)
+		return (ret);
+	}
 
-	/* Go with update */
-	spin_lock_irqsave(&ha->hardware_lock, cpu_flags);
-	status = qla2x00_set_flash_image(ha, kern_tmp);
-	spin_unlock_irqrestore(&ha->hardware_lock, cpu_flags);
-
-	/* Schedule DPC to restart the RISC */
-	set_bit(ISP_ABORT_NEEDED, &ha->dpc_flags);
-	up(ha->dpc_wait);
+	status = qla2x00_update_or_read_flash(ha, kern_tmp, 0, 
+			FLASH_IMAGE_SIZE, QLA2X00_WRITE); 
 
 	KMEM_FREE(kern_tmp, FLASH_IMAGE_SIZE);
 
@@ -629,3 +731,201 @@ int qla2x00_update_option_rom(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
 
 	return (ret);
 }
+
+int
+qla2x00_update_option_rom_ext(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
+{
+	int		ret;
+	uint8_t		*usr_tmp;
+	uint8_t		*kern_tmp;
+	uint16_t	status;
+	int		iter, found;
+	uint32_t	saddr, length;
+
+	DEBUG9(printk("%s: entered.\n", __func__);)
+
+	found = 0;
+	saddr = length = 0;
+	/* Retrieve region or raw starting address. */
+	if (pext->SubCode == 0xFFFF) {
+		saddr = pext->Reserved1;
+		length = pext->RequestLen;
+		found++;
+	} else {
+		INT_OPT_ROM_REGION *OptionRomTable = NULL;
+		unsigned long  OptionRomTableSize;
+		/* Pick the right OptionRom table based on device id */
+		qla2x00_get_option_rom_table(ha, &OptionRomTable, 
+		    	&OptionRomTableSize);
+		for (iter = 0; OptionRomTable != NULL && iter <
+		    (OptionRomTableSize / sizeof(INT_OPT_ROM_REGION));
+		    iter++) {
+			if (OptionRomTable[iter].Region == pext->SubCode) {
+				saddr = OptionRomTable[iter].Beg;
+				length = OptionRomTable[iter].Size;
+				found++;
+				break;
+			}
+		}
+	}
+	if (!found) {
+		pext->Status = EXT_STATUS_ERR;
+		return 1;
+	}
+	if (pext->RequestLen < length) {
+		pext->Status = EXT_STATUS_COPY_ERR;
+		return 1;
+	}
+
+	pext->Status = EXT_STATUS_OK;
+	pext->DetailStatus = EXT_STATUS_OK;
+
+	/* Read from user buffer */
+	usr_tmp = (uint8_t *)pext->RequestAdr;
+
+	kern_tmp = (uint8_t *)KMEM_ZALLOC(length, 30);
+	if (kern_tmp == NULL) {
+		pext->Status = EXT_STATUS_COPY_ERR;
+		printk(KERN_WARNING
+		    "%s: ERROR in flash allocation.\n", __func__);
+		return 1;
+	}
+	ret = copy_from_user(kern_tmp, usr_tmp, length);
+	if (ret) {
+		KMEM_FREE(kern_tmp, length);
+		pext->Status = EXT_STATUS_COPY_ERR;
+		DEBUG9_10(printk("%s: ERROR in buffer copy READ. "
+		    "RequestAdr=%p\n", __func__, pext->RequestAdr));
+		return (ret);
+	}
+
+	/* Go with update */
+	status = qla2x00_update_or_read_flash(ha, kern_tmp, saddr, 
+	    		length, QLA2X00_WRITE);
+
+	KMEM_FREE(kern_tmp, length);
+
+	if (status) {
+		ret = 1;
+		pext->Status = EXT_STATUS_COPY_ERR;
+		DEBUG9_10(printk("%s: ERROR updating flash.\n", __func__);)
+	}
+
+	DEBUG9(printk("%s: exiting.\n", __func__);)
+	return (ret);
+}
+
+int
+qla2x00_get_option_rom_layout(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
+{
+	int	ret, iter;
+	INT_OPT_ROM_REGION *OptionRomTable = NULL;
+	INT_OPT_ROM_LAYOUT *optrom_layout;	
+	unsigned long  OptionRomTableSize; 
+
+	DEBUG9(printk("%s: entered.\n", __func__);)
+
+	/* Pick the right OptionRom table based on device id */
+	qla2x00_get_option_rom_table(ha, &OptionRomTable, &OptionRomTableSize);
+
+	if (OptionRomTable == NULL) {
+		pext->Status = EXT_STATUS_DEV_NOT_FOUND;
+		DEBUG9_10(printk("%s(%ld) Option Rom Table for device_id=0x%x "
+		    "not defined\n", __func__,ha->host_no,ha->device_id));
+		return pext->Status;
+	}
+
+	if (pext->ResponseLen < OptionRomTableSize) {
+		pext->Status = EXT_STATUS_BUFFER_TOO_SMALL;
+		DEBUG9_10(printk("%s(%ld) buffer too small: response_len = %d "
+		    "optrom_table_len=%ld.\n", __func__, ha->host_no,
+		    pext->ResponseLen,OptionRomTableSize));
+		return pext->Status;
+	}
+	if (qla2x00_get_ioctl_scrap_mem(ha, (void **)&optrom_layout,
+	    OptionRomTableSize)) {
+		/* not enough memory */
+		pext->Status = EXT_STATUS_NO_MEMORY;
+		DEBUG9_10(printk("%s(%ld): inst=%ld scrap not big enough. "
+		    "size requested=%ld.\n", __func__, ha->host_no,
+		    ha->instance, OptionRomTableSize));
+		return pext->Status;
+	}
+
+	// Dont Count the NULL Entry.
+	optrom_layout->NoOfRegions =
+	    (OptionRomTableSize / sizeof(INT_OPT_ROM_REGION) - 1);
+
+	for (iter = 0; iter < optrom_layout->NoOfRegions; iter++) {
+		optrom_layout->Region[iter].Region =
+		    OptionRomTable[iter].Region;
+		optrom_layout->Region[iter].Size =
+		    OptionRomTable[iter].Size;
+
+		if (OptionRomTable[iter].Region == INT_OPT_ROM_REGION_ALL)
+			optrom_layout->Size = OptionRomTable[iter].Size;
+	}
+
+	ret = copy_to_user(pext->ResponseAdr, optrom_layout,
+	    		OptionRomTableSize);
+	if (ret) {
+		pext->Status = EXT_STATUS_COPY_ERR;
+		DEBUG9_10(printk("%s(%ld): inst=%ld ERROR copy rsp buffer.\n",
+		    __func__, ha->host_no, ha->instance));
+		qla2x00_free_ioctl_scrap_mem(ha);
+		return ret;
+	}
+
+	pext->Status       = EXT_STATUS_OK;
+	pext->DetailStatus = EXT_STATUS_OK;
+
+	qla2x00_free_ioctl_scrap_mem(ha);
+
+	DEBUG9(printk("%s: exiting.\n", __func__));
+	return ret;
+}
+
+
+/*
+* qla2x00_get_option_rom_table 
+* 	This function returns the OptionRom Table for matching device id.
+*/ 	
+static void
+qla2x00_get_option_rom_table(scsi_qla_host_t *ha,
+    INT_OPT_ROM_REGION **pOptionRomTable, unsigned long  *OptionRomTableSize)
+{
+	DEBUG9(printk("%s: entered.\n", __func__));
+
+	switch (ha->device_id) {
+	case QLA6312_DEVICE_ID:
+		*pOptionRomTable = OptionRomTable6312;
+		*OptionRomTableSize = sizeof(OptionRomTable6312);
+		break;		       	
+	case QLA2312_DEVICE_ID:
+		/* HBA Model 6826A - is 2312 V3 Chip */
+		if (ha->subsystem_vendor == 0x103C &&
+		    ha->subsystem_device == 0x12BA){
+			*pOptionRomTable = OptionRomTable6826A;
+			*OptionRomTableSize = sizeof(OptionRomTable6826A);
+		} else {
+			*pOptionRomTable = OptionRomTable2312;
+			*OptionRomTableSize = sizeof(OptionRomTable2312);
+		}
+		break;
+	case QLA2322_DEVICE_ID:
+	case QLA6322_DEVICE_ID:
+		*pOptionRomTable = OptionRomTable2322;
+		*OptionRomTableSize = sizeof(OptionRomTable2322);
+		break;
+	case QLA2400_DEVICE_ID:
+		*pOptionRomTable = OptionRomTable2400;
+		*OptionRomTableSize = sizeof(OptionRomTable2400);
+		break;
+	default: 
+		DEBUG9_10(printk("%s(%ld) Option Rom Table for device_id=0x%x "
+		    "not defined\n", __func__, ha->host_no,ha->device_id));
+		break;
+	}
+
+	DEBUG9(printk("%s: exiting.\n", __func__);)
+}
